"""
Implement the basic BMC for a symbolic transition system

To explain:
For using Boolean logic in PySMT: https://pysmt.readthedocs.io/en/stable/tutorials/boolean_logic.html
To accumulate conjunctions

"""

from pysmt.shortcuts import *



def get_next(v):
    """
    Given v, returns v'
    """
    return Symbol("%s_next" % v.symbol_name(), v.symbol_type())

def get_nexts(V):
    """
    Given v, returns v'
    """
    return [Symbol("%s_next" % v.symbol_name(), v.symbol_type()) for v in V]



class SymbolicTS():
    def __init__(self, V, init, trans):
        """
        Symbolic transition system:
        V: set of Boolean variables (Represented as a Symbol from PySMT)
        I: initial states, formula only containing variable from V (PySMT formula)
        T: transition relation, formula only containing variable from V and V', here called V_next (PySMT formula)

        About V_next:
        We assume we can map a variables from V to V_next just by modifying the name of v.
        Given a variable "a" you get the primed variable named a_next.

        This means that your initial set of variables should not already terminate
        with the suffix _next (otherwise, there could be some variable capture...)

        Given a variable, you can get the corresponding next variable with the function
        get_next.

        """
        self.V = set([v for v in V])
        self.V_next = set([get_next(v) for v in V])
        self.init = init
        self.trans = trans


    def retime(self, formula, time):
        """
        Given a formula containing the variables V and V_next and an integer >= 0,
        returns a formula where:
        - all the V-s variables have been replaced with v_at_time variables, and
        - all the V_next-s variables v_next have been replaced with v_at_<time+1>.

        Example
        Assume V = {a,b,c}, V_next = {a_next, b_next, c_next}, and formula is
        the formula "(a & b & c) | !c')".

        self.retime(formula, 2) returns the formula:
        "(a_at_2 & b_at_2 & c_at_2) | !c_at_3)".
        """

        def at_time(v, time, is_next = False):
            assert time >= 0

            if not is_next:
                assert not v.symbol_name().endswith("_next")
                symb_name = "%s_at_%d" % (v.symbol_name(),time) 
            else:
                assert v.symbol_name().endswith("_next")
                name = v.symbol_name().removesuffix("_next")
                symb_name = "%s_at_%d" % (name,time)
            return Symbol(symb_name, v.symbol_type())

        f1 = formula.substitute({ a : at_time(a,time) for a in self.V} )
        f2 = f1.substitute({ a : at_time(a,time+1, True) for a in self.V_next} )

        return f2

def print_cex(lasso_cex):
    """ Print a counter-example.
    """
    cex,loopback = lasso_cex
    has_lasso = loopback >= 0

    assert(( not has_lasso) or set(cex[-1]) == set(cex[loopback])) # wrong loopback

    kind = "Lasso-shaped" if has_lasso else "Non lasso-shaped"
    print(kind, "counter example of length", len(cex) - 1)

    for i in range(len(cex)):
        print("State %d.%s" % (i,
                               "" if loopback != i else " The loop starts here." ))
        for (var,value) in cex[i]:
            print("  %s = %s" % (var, str(value)))


def check_formula(ts, prop, encoding, k):
    """
    Performs the satisfiability check of the formula encoding and prints
    a counterexample if one is found.
    """
    is_true = not is_sat(encoding)
    if not is_true:
        print(prop, "is false after",  k, "stesps!")

        cex = bmc_get_cex(ts, encoding, k)
        if cex is not None:
            print_cex(cex)
    else:
        print(prop, "is unknown after", k, "steps.")
    print("")
    return is_true



def bmc_Gp_no_loop(ts, p, k):
    """
    Find a violations to the property G p exploring k steps

    Here, don't care about loopback (find a finite-path)

    Here, you have to create the BMC formula that finds a violation to
    properties of the form G p.
    Create the BMC formula in the encoding variable.

    """

    encoding = None

    if encoding is None:
        raise NotImplementedError()

    return check_formula(ts, "G " + p.serialize(), encoding, k)



def bmc_get_cex(ts, encoding, k):
    """ This function creates a counter-example (a path in the transition
    system) from the model found by the SAT solver.

    We expect the counter example to be a list of lists.
    Each element of the counter-example list is a list of assignments to variables
    of the transition systems. We represent an assignment as a tuple (e.g., (a,True) 
    means that the variable a is assigned to the boolean value True).
    Valid assignments either assign the variable to True or False.

    So, for example, a counter-example with 3 states would look like:

    [ [(a,True), (b, False)], [(a,True), (b, True)] [(a,True), (b, False)] ]

    Here, you also need to check if there is a loop from the last state k
    to one of the previous states.
    The function will return the index of the loop, if one exist, or -1 otherwise.

    In the end, the function returns a tuple containing the counterexample list 
    and the integer number for the loopback.

    HINT: you can get the value assigned to a variable in a model m with the function 
    get_py_value (e.g., m.get_py_value(variable) returns the value, either True or False,
    the SAT solver assigned to variable).

    """

    # Gets a model from the SAT solver
    m = get_model(encoding)
    assert m is not None

    # Process the model and creates 

    return None

def bmc_Gp(ts, p, k):
    """
    Find a violations to the property G p exploring k steps

    Here you will care about finding an infinite path.
    """

    encoding = None

    if encoding is None:
        raise NotImplementedError()

    return check_formula(ts, "G " + p.serialize(), encoding, k)


def bmc_Fp(ts, p, k):
    """
    Find a violations to the property F p exploring k steps
    """

    encoding = None

    if encoding is None:
        raise NotImplementedError()

    return check_formula(ts, "F " + p.serialize(), encoding, k)





def bmc_FGp(ts, p, k):
    """
    Find a violations to the property FG p exploring k steps
    """

    encoding = None

    if encoding is None:
        raise NotImplementedError()

    return check_formula(ts, "F G " + p.serialize(), encoding, k)
