""" Transition System

A transition system is the tuple: <states, actions, tr, init, ap, labels>

A state is an integer number of a tuple of states (inductive definition)
An action is a string
An atomic proposition (Prop) is a string
A transition is a tuple in State x Action x State

Example of use:

# Creates the states
s1 = (1,)
s2 = (2,)

# Creates a transition system
t = TS(set([s1, s2]), set(["a","b"]),
       set([(s1,"a",s1),(s2,"b",s1)]),
       set([s1]),
       set(["init","final"]),
       {s1 : set(["init"]), s2 : set()})
"""

from __future__ import annotations
from collections.abc import Mapping, Iterator, Iterable
from typing import TypeAlias, Type, Tuple, Callable, Optional, NewType, TypeVar, Union, TypedDict
from copy import deepcopy

from mc308.data_structures.formula import PropFormula
from pysmt.shortcuts import FALSE, TRUE, Symbol

from functools import reduce, partial
import itertools

class StatesTuple(tuple):
    """ Tuples of states are useful to represent products """
    def __new__(self, *args : int | StatesTuple):
        return tuple.__new__(StatesTuple, args)


States =  int | StatesTuple
Action = NewType('Action', str)
Prop = NewType('Prop', str)
Trans = Tuple['States','Action','States']

class TS():
    def __init__(self : TS,
                 states : frozenset[States],
                 actions : frozenset[Action],
                 tr : frozenset[Trans],
                 init : frozenset[States],
                 ap : frozenset[Prop],
                 labels : Mapping[States,frozenset[Prop]]) -> None:
        """
        Creates a finite state transition system
        """

        self.states = set([a for a in states])
        self.actions = set([a for a in actions])
        self.tr = set([t for t in tr])
        self.init = set([i for i in init])
        self.ap = set([i for i in ap])
        self.labels = {}
        for k,v in labels.items():
            self.labels[k] = set([el for el in v])

    def post(self, s, act=None):
        res = []
        for (src, a, dst) in self.tr:
            if src == s and (act is None or act == a):
                res.append((src,a,dst))
        return res

    def satisfy(self : TS, s : States, formula : PropFormula) -> bool:
        """
        Return true iff s |= formula

        TODO: add caching for model construction and for caching
        evaluation
        """

        subs = {}
        for v in formula.get_free_variables():
            subs[v] = FALSE()
        for l in self.ap:
            subs[Symbol(l)] = FALSE()
        for l in self.labels[s]:
            subs[Symbol(l)] = TRUE()
        return (TRUE() == formula.substitute(subs).simplify())

    def simulate(self : TS, run : List[States]) -> Tuple[bool,Option[List[T]]]:
        """
        Simulates an infinite run of TS, returns true and the correspondent trace,
        false otherwise.
        """
        trace = []
        if len(run) <= 0: return False, None

        # Check init
        if not run[0] in self.init:
            return False, None
        trace.append(self.labels[run[0]])

        # check trans
        i = 0
        has_loop = False
        while (i + 1 < len(run)):
            has_loop = has_loop or run[i] == run[-1]

            found = False
            for (_, a, succ) in self.post(run[i]):
                if succ == run[i+1]:
                    trace.append(self.labels[succ])
                    found = True
                    break

            if not found:
                return False, None

            i += 1

        if not has_loop:
            return False, None

        return True, trace

    def __eq__(self : TS, other):
        """
        Warning: shallow equality.
        """
        if isinstance(other, TS):
            return (self.states == other.states and
                    self.actions == other.actions and
                    self.tr == other.tr and
                    self.init == other.init and
                    self.ap == other.ap and
                    self.labels == other.labels)
        else:
            return False

    def __repr__(self : TS) :
        # TODO use helper function with streams

        pl = lambda l : ",".join(l)

        return """TS : {
states : {%s};
  actions : {%s};
  tr : {%s};
  init : {%s};
  ap : {%s};
  labels : {%s};
}""" % ( pl([str(l) for l in self.states]),
         pl(self.actions),
         pl(["(%s, %s, %s)" % (str(a[0]), a[1], str(a[2])) for a in self.tr]),
         pl([str(l) for l in self.init]),
         pl(self.ap),
         pl(["%s : {%s}" % (str(a[0]),pl(a[1])) for a in self.labels.items()]))


################################################################################
# Utils
def state_prod(s1 : States, s2 : States) -> States:
    """
    Compute the product of two states.
    """
    prod = StatesTuple(s1,s2)
    return prod


################################################################################
# TS operations

def trans_op(t1 : Trans, t2 : Trans) -> Optional[Trans]:
    """
    Returns None if t1 and t2 do not move on the same actions,
    otherwise returns the product.
    """

    (s1, a1, d1) = t1
    (s2, a2, d2) = t2

    trans = None if a1 != a2 else (state_prod(s1, s2),
                                   a1,
                                   state_prod(d1, d2))
    return trans


T = TypeVar('T')
def cart_prod(e1 : frozenset[T], e2 : frozenset[T],
              op : Callable[[T,T],  Optional[T]]) -> frozenset[T] :
    """
    Computes the cartesian product of the set e1 and and e2.
    """
    res = set()
    for l1 in e1:
        for l2 in e2:
            r = op(l1,l2)
            if (None != r):
                res.add(r)
    return frozenset(res)



def interleave_generic(ts1 : TS, ts2 : TS,
                       common : Optional[Iterable[Action]]) -> TS:
    """
    Computes the interleaving of two transition systems.
    """
    # compute the product only for the reachable states
    init_states = cart_prod(ts1.init, ts2.init, state_prod)
    trans = set()

    if common is None:
        common = []

    states = set()
    for s in init_states:
        stack = [s]

        while (0 < len(stack)):
            s = stack.pop()
            if s in states:
                continue

            states.add(s)

            s1,s2 = s
            trans_on_common = {}
            for c in common:
                trans_on_common[c] = set()

            for (s1, a1, d1) in ts1.post(s1):
                if not a1 in common:
                    dst_state = state_prod(d1, s2)
                    trans.add((s,a1,dst_state))
                    stack.append(dst_state)
                else:
                    trans_on_common[a1].add((s1, a1, d1))

            for (s2, a2, d2) in ts2.post(s2):
                if not a2 in common:
                    dst_state = state_prod(s1, d2)
                    trans.add((s,a2,dst_state))
                    stack.append(dst_state)
                else:
                    for (s1,a1,d1) in trans_on_common[a2]:
                        assert(a1 == a2)
                        dst_state = state_prod(d1, d2)
                        trans.add((s,a2,dst_state))
                        stack.append(dst_state)

    def build_labels(states):
        """
        rebuild the labels for the states in the product
        """
        base : dict[States, frozenset[Prop]] = {}
        for s in states:
            base[s] = ts1.labels[s[0]].union(ts2.labels[s[1]])
        return base

    return TS(states,
              ts1.actions.union(ts2.actions),
              trans,
              init_states,
              ts1.ap.union(ts2.ap),
              build_labels(states)
              )

def interleave(ts1 : TS, ts2 : TS):
    return interleave_generic(ts1,ts2,set())


def handshake(ts1 : TS, ts2 : TS, H : frozenset[Action]) -> TS:
    """
    Compute the interleaving of two transition systems.

    ts1 and ts2:
      - have to synchronize on the same actions
      - move independently on the other actions
    """
    return interleave_generic(ts1,ts2,H)





